# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
import logging
import sys

import pika

from swh.core.statsd import statsd
from swh.scheduler import get_scheduler

logger = logging.getLogger(__name__)


def utcnow():
    return datetime.datetime.now(tz=datetime.timezone.utc)


def get_listener(broker_url, queue_name, scheduler_backend):
    connection = pika.BlockingConnection(
        pika.URLParameters(broker_url)
    )
    channel = connection.channel()

    channel.queue_declare(queue=queue_name, durable=True)

    exchange = 'celeryev'
    routing_key = '#'
    channel.queue_bind(queue=queue_name, exchange=exchange,
                       routing_key=routing_key)

    channel.basic_qos(prefetch_count=1000)

    channel.basic_consume(
        queue=queue_name,
        on_message_callback=get_on_message(scheduler_backend),
    )

    return channel


def get_on_message(scheduler_backend):
    def on_message(channel, method_frame, properties, body):
        try:
            events = json.loads(body)
        except Exception:
            logger.warning('Could not parse body %r', body)
            events = []

        if not isinstance(events, list):
            events = [events]

        for event in events:
            logger.debug('Received event %r', event)
            process_event(event, scheduler_backend)

        channel.basic_ack(delivery_tag=method_frame.delivery_tag)

    return on_message


def process_event(event, scheduler_backend):
    uuid = event.get('uuid')
    if not uuid:
        return

    event_type = event['type']
    statsd.increment('swh_scheduler_listener_handled_event_total',
                     tags={'event_type': event_type})

    if event_type == 'task-started':
        scheduler_backend.start_task_run(
            uuid, timestamp=utcnow(),
            metadata={'worker': event.get('hostname')},
        )
    elif event_type == 'task-result':
        result = event['result']

        status = None

        if isinstance(result, dict) and 'status' in result:
            status = result['status']
            if status == 'success':
                status = 'eventful' if result.get('eventful') else 'uneventful'

        if status is None:
            status = 'eventful' if result else 'uneventful'

        scheduler_backend.end_task_run(uuid, timestamp=utcnow(),
                                       status=status, result=result)
    elif event_type == 'task-failed':
        scheduler_backend.end_task_run(uuid, timestamp=utcnow(),
                                       status='failed')


if __name__ == '__main__':
    url = sys.argv[1]
    logging.basicConfig(level=logging.DEBUG)
    scheduler_backend = get_scheduler('local', args={
        'db': 'service=swh-scheduler'
    })
    channel = get_listener(url, 'celeryev.test', scheduler_backend)
    logger.info('Start consuming')
    channel.start_consuming()



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import traceback
import yaml
import pkg_resources

from click.testing import CliRunner

from swh.scheduler import get_scheduler
from swh.scheduler.cli import cli


FAKE_MODULE_ENTRY_POINTS = {
    'lister.gnu=swh.lister.gnu:register',
    'lister.pypi=swh.lister.pypi:register',
}


@pytest.fixture
def mock_pkg_resources(monkeypatch):
    """Monkey patch swh.scheduler's mock_pkg_resources.iter_entry_point call

    """
    def fake_iter_entry_points(*args, **kwargs):
        """Substitute fake function to return a fixed set of entrypoints

        """
        from pkg_resources import EntryPoint, Distribution
        d = Distribution()
        return [EntryPoint.parse(entry, dist=d)
                for entry in FAKE_MODULE_ENTRY_POINTS]

    original_method = pkg_resources.iter_entry_points
    monkeypatch.setattr(
        pkg_resources, "iter_entry_points", fake_iter_entry_points)

    yield

    # reset monkeypatch: is that needed?
    monkeypatch.setattr(
        pkg_resources, "iter_entry_points", original_method)


@pytest.fixture
def local_sched_config(swh_scheduler_config):
    """Expose the local scheduler configuration

    """
    return {
        'scheduler': {
            'cls': 'local',
            'args': swh_scheduler_config
        }
    }


@pytest.fixture
def local_sched_configfile(local_sched_config, tmp_path):
    """Write in temporary location the local scheduler configuration

    """
    configfile = tmp_path / 'config.yml'
    configfile.write_text(yaml.dump(local_sched_config))
    return configfile.as_posix()


def test_register_ttypes_all(
        mock_pkg_resources, local_sched_config, local_sched_configfile):
    """Registering all task types"""

    for command in [
        ['--config-file', local_sched_configfile, 'task-type', 'register'],
        ['--config-file', local_sched_configfile, 'task-type', 'register',
         '-p', 'all'],
        ['--config-file', local_sched_configfile, 'task-type', 'register',
         '-p', 'lister.gnu',
         '-p', 'lister.pypi'],
    ]:
        result = CliRunner().invoke(cli, command)

        assert result.exit_code == 0, traceback.print_exception(
            *result.exc_info)

        scheduler = get_scheduler(**local_sched_config['scheduler'])
        all_tasks = [
            'list-gnu-full',
            'list-pypi',
        ]
        for task in all_tasks:
            task_type_desc = scheduler.get_task_type(task)
            assert task_type_desc
            assert task_type_desc['type'] == task
            assert task_type_desc['backoff_factor'] == 1


def test_register_ttypes_filter(
        mock_pkg_resources, local_sched_config, local_sched_configfile):
    """Filtering on one worker should only register its associated task type

    """
    result = CliRunner().invoke(cli, [
        '--config-file', local_sched_configfile,
        'task-type', 'register', '--plugins', 'lister.gnu'
    ])

    assert result.exit_code == 0, traceback.print_exception(*result.exc_info)

    scheduler = get_scheduler(**local_sched_config['scheduler'])
    all_tasks = [
        'list-gnu-full',
    ]
    for task in all_tasks:
        task_type_desc = scheduler.get_task_type(task)
        assert task_type_desc
        assert task_type_desc['type'] == task
        assert task_type_desc['backoff_factor'] == 1



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import time

import click

from . import cli


@cli.command('start-runner')
@click.option('--period', '-p', default=0,
              help=('Period (in s) at witch pending tasks are checked and '
                    'executed. Set to 0 (default) for a one shot.'))
@click.pass_context
def runner(ctx, period):
    """Starts a swh-scheduler runner service.

    This process is responsible for checking for ready-to-run tasks and
    schedule them."""
    from swh.scheduler.celery_backend.runner import run_ready_tasks
    from swh.scheduler.celery_backend.config import build_app

    app = build_app(ctx.obj['config'].get('celery'))
    app.set_current()

    logger = logging.getLogger(__name__ + '.runner')
    scheduler = ctx.obj['scheduler']
    logger.debug('Scheduler %s' % scheduler)
    try:
        while True:
            logger.debug('Run ready tasks')
            try:
                ntasks = len(run_ready_tasks(scheduler, app))
                if ntasks:
                    logger.info('Scheduled %s tasks', ntasks)
            except Exception:
                logger.exception('Unexpected error in run_ready_tasks()')
            if not period:
                break
            time.sleep(period)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command('start-listener')
@click.pass_context
def listener(ctx):
    """Starts a swh-scheduler listener service.

    This service is responsible for listening at task lifecycle events and
    handle their workflow status in the database."""
    scheduler_backend = ctx.obj['scheduler']
    if not scheduler_backend:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    broker = ctx.obj['config']\
                .get('celery', {})\
                .get('task_broker', 'amqp://guest@localhost/%2f')

    from swh.scheduler.celery_backend.pika_listener import get_listener

    listener = get_listener(broker, 'celeryev.listener', scheduler_backend)
    try:
        listener.start_consuming()
    finally:
        listener.stop_consuming()


@cli.command('rpc-serve')
@click.option('--host', default='0.0.0.0',
              help="Host to run the scheduler server api")
@click.option('--port', default=5008, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=None,
              help=("Indicates if the server should run in debug mode. "
                    "Defaults to True if log-level is DEBUG, False otherwise.")
              )
@click.pass_context
def rpc_server(ctx, host, port, debug):
    """Starts a swh-scheduler API HTTP server.
    """
    if ctx.obj['config']['scheduler']['cls'] == 'remote':
        click.echo("The API server can only be started with a 'local' "
                   "configuration", err=True)
        ctx.exit(1)

    from swh.scheduler.api import server
    server.app.config.update(ctx.obj['config'])
    if debug is None:
        debug = ctx.obj['log_level'] <= logging.DEBUG
    server.app.run(host, port=port, debug=bool(debug))



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Elastic Search backend

"""

import datetime  # noqa
import logging

from copy import deepcopy
from typing import Any, Dict

from elasticsearch import helpers

from swh.core import utils

logger = logging.getLogger(__name__)


DEFAULT_CONFIG = {
    'elasticsearch': {
        'cls': 'local',
        'args': {
            'index_name_prefix': 'swh-tasks',
            'storage_nodes': ['localhost:9200'],
            'client_options': {
                'sniff_on_start': False,
                'sniff_on_connection_fail': True,
                'http_compress': False,
                'sniffer_timeout': 60
            },
        },
    }
}


def get_elasticsearch(cls: str, args: Dict[str, Any] = {}):
    """Instantiate an elastic search instance

    """
    if cls == 'local':
        from elasticsearch import Elasticsearch
    elif cls == 'memory':
        from .elasticsearch_memory import MemoryElasticsearch as Elasticsearch  # type: ignore  # noqa
    else:
        raise ValueError('Unknown elasticsearch class `%s`' % cls)

    return Elasticsearch(**args)


class ElasticSearchBackend:
    """ElasticSearch backend to index tasks

    This uses an elasticsearch client to actually discuss with the
    elasticsearch instance.

    """
    def __init__(self, **config):
        self.config = deepcopy(DEFAULT_CONFIG)
        self.config.update(config)
        es_conf = self.config['elasticsearch']
        args = deepcopy(es_conf['args'])
        self.index_name_prefix = args.pop('index_name_prefix')
        self.storage = get_elasticsearch(
            cls=es_conf['cls'],
            args={
                'hosts': args.get('storage_nodes', []),
                **args.get('client_options', {}),
            }
        )
        # document's index type (cf. /data/elastic-template.json)
        self.doc_type = 'task'

    def initialize(self):
        self.storage.indices.put_mapping(
            index=f"{self.index_name_prefix}-*",
            doc_type=self.doc_type,
            # to allow type definition below
            include_type_name=True,
            # to allow install mapping even if no index yet
            allow_no_indices=True,
            body={
                "properties": {
                    "task_id": {"type": "double"},
                    "task_policy": {"type": "text"},
                    "task_status": {"type": "text"},
                    "task_run_id": {"type": "double"},
                    "arguments": {
                        "type": "object",
                        "properties": {
                            "args": {
                                "type": "nested",
                                "dynamic": False
                            },
                            "kwargs": {
                                "type": "text"
                            }
                        }
                    },
                    "type": {"type": "text"},
                    "backend_id": {"type": "text"},
                    "metadata": {
                        "type": "object",
                        "enabled": False
                    },
                    "scheduled":  {
                        "type": "date",
                        "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||strict_date_optional_time||epoch_millis"  # noqa
                    },
                    "started":  {
                        "type": "date",
                        "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||strict_date_optional_time||epoch_millis"  # noqa
                    },
                    "ended":  {
                        "type": "date",
                        "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||strict_date_optional_time||epoch_millis"  # noqa
                    }
                }
            })

        self.storage.indices.put_settings(
            index=f"{self.index_name_prefix}-*",
            allow_no_indices=True,
            body={
                "index": {
                    "codec": "best_compression",
                    "refresh_interval": "1s",
                    "number_of_shards": 1
                }
            })

    def create(self, index_name) -> None:
        """Create and initialize index_name with mapping for all indices
           matching `swh-tasks-` pattern

        """
        assert index_name.startswith(self.index_name_prefix)
        self.storage.indices.create(index_name)

    def compute_index_name(self, year, month):
        """Given a year, month, compute the index's name.

        """
        return '%s-%s-%s' % (
            self.index_name_prefix, year, '%02d' % month)

    def mget(self, index_name, doc_ids, chunk_size=500,
             source=True):
        """Retrieve document's full content according to their ids as per
           source's setup.

           The `source` allows to retrieve only what's interesting, e.g:
           - source=True ; gives back the original indexed data
           - source=False ; returns without the original _source field
           - source=['task_id'] ; returns only task_id in the _source field

        Args:
            index_name (str): Name of the concerned index.
            doc_ids (generator): Generator of ids to retrieve
            chunk_size (int): Number of documents chunk to send for retrieval
            source (bool/[str]): Source of information to return

        Yields:
            document indexed as per source's setup

        """
        if isinstance(source, list):
            source = {'_source': ','.join(source)}
        else:
            source = {'_source': str(source).lower()}

        for ids in utils.grouper(doc_ids, n=1000):
            res = self.storage.mget(body={'ids': list(ids)},
                                    index=index_name,
                                    doc_type=self.doc_type,
                                    params=source)
            if not res:
                logger.error('Error during retrieval of data, skipping!')
                continue

            for doc in res['docs']:
                found = doc.get('found')
                if not found:
                    msg = 'Doc id %s not found, not indexed yet' % doc['_id']
                    logger.warning(msg)
                    continue
                yield doc['_source']

    def _streaming_bulk(self, index_name, doc_stream, chunk_size=500):
        """Bulk index data and returns the successful indexed data's
           identifier.

        Args:
            index_name (str): Name of the concerned index.
            doc_stream (generator): Generator of documents to index
            chunk_size (int): Number of documents chunk to send for indexation

        Yields:
            document id indexed

        """
        actions = ({'_index': index_name,
                    '_op_type': 'index',
                    '_type': self.doc_type,
                    '_source': data} for data in doc_stream)
        for ok, result in helpers.streaming_bulk(client=self.storage,
                                                 actions=actions,
                                                 chunk_size=chunk_size,
                                                 raise_on_error=False,
                                                 raise_on_exception=False):
            if not ok:
                logger.error('Error during %s indexation. Skipping.', result)
                continue
            yield result['index']['_id']

    def is_index_opened(self, index_name: str) -> bool:
        """Determine if an index is opened or not

        """
        try:
            self.storage.indices.stats(index_name)
            return True
        except Exception:
            # fails when indice is closed (no other api call found)
            return False

    def streaming_bulk(self, index_name, doc_stream, chunk_size=500,
                       source=True):
        """Bulk index data and returns the successful indexed data as per
           source's setup.

           the `source` permits to retrieve only what's of interest to
           us, e.g:

           - source=True ; gives back the original indexed data
           - source=False ; returns without the original _source field
           - source=['task_id'] ; returns only task_id in the _source field

           Note that:
           - if the index is closed, it will be opened
           - if the index does not exist, it will be created and opened

           This keeps the index opened for performance reasons.

        Args:
            index_name (str): Name of the concerned index.
            doc_stream (generator): Document generator to index
            chunk_size (int): Number of documents chunk to send
            source (bool, [str]): the information to return

        """
        # index must exist
        if not self.storage.indices.exists(index_name):
            self.create(index_name)
        # index must be opened
        if not self.is_index_opened(index_name):
            self.storage.indices.open(index_name)

        indexed_ids = self._streaming_bulk(
            index_name, doc_stream, chunk_size=chunk_size)
        yield from self.mget(
            index_name, indexed_ids, chunk_size=chunk_size, source=source)



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup


@click.group(name='scheduler', context_settings=CONTEXT_SETTINGS,
             cls=AliasedGroup)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.option('--database', '-d', default=None,
              help="Scheduling database DSN (imply cls is 'local')")
@click.option('--url', '-u', default=None,
              help="Scheduler's url access (imply cls is 'remote')")
@click.option('--no-stdout', is_flag=True, default=False,
              help="Do NOT output logs on the console")
@click.pass_context
def cli(ctx, config_file, database, url, no_stdout):
    """Software Heritage Scheduler tools.

    Use a local scheduler instance by default (plugged to the
    main scheduler db).
    """
    from swh.core import config
    from swh.scheduler.celery_backend.config import setup_log_handler
    from swh.scheduler import get_scheduler, DEFAULT_CONFIG

    ctx.ensure_object(dict)
    log_level = ctx.obj.get('log_level', logging.INFO)

    setup_log_handler(
        loglevel=log_level, colorize=False,
        format='[%(levelname)s] %(name)s -- %(message)s',
        log_console=not no_stdout)

    logger = logging.getLogger(__name__)
    scheduler = None
    conf = config.read(config_file, DEFAULT_CONFIG)
    if 'scheduler' not in conf:
        raise ValueError("missing 'scheduler' configuration")

    if database:
        conf['scheduler']['cls'] = 'local'
        conf['scheduler']['args']['db'] = database
    elif url:
        conf['scheduler']['cls'] = 'remote'
        conf['scheduler']['args'] = {'url': url}
    sched_conf = conf['scheduler']
    try:
        logger.debug('Instanciating scheduler with %s' % (
            sched_conf))
        scheduler = get_scheduler(**sched_conf)
    except ValueError:
        # it's the subcommand to decide whether not having a proper
        # scheduler instance is a problem.
        pass

    ctx.obj['scheduler'] = scheduler
    ctx.obj['config'] = conf

from . import admin, task, task_type  # noqa


def main():
    import click.core
    click.core.DEPRECATED_HELP_NOTICE = '''

DEPRECATED! Please use the command 'swh scheduler'.'''
    cli.deprecated = True
    return cli(auto_envvar_prefix='SWH_SCHEDULER')


if __name__ == '__main__':
    main()



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime


TEMPLATES = {
    'git': {
        'type': 'update-git',
        'arguments': {
            'args': [],
            'kwargs': {},
        },
        'next_run': None,
    },
    'hg': {
        'type': 'update-hg',
        'arguments': {
            'args': [],
            'kwargs': {},
        },
        'next_run': None,
        'policy': 'oneshot',
    }
}


TASK_TYPES = {
    'git': {
        'type': 'update-git',
        'description': 'Update a git repository',
        'backend_name': 'swh.loader.git.tasks.UpdateGitRepository',
        'default_interval': datetime.timedelta(days=64),
        'min_interval': datetime.timedelta(hours=12),
        'max_interval': datetime.timedelta(days=64),
        'backoff_factor': 2,
        'max_queue_length': None,
        'num_retries': 7,
        'retry_delay': datetime.timedelta(hours=2),
    },
    'hg': {
        'type': 'update-hg',
        'description': 'Update a mercurial repository',
        'backend_name': 'swh.loader.mercurial.tasks.UpdateHgRepository',
        'default_interval': datetime.timedelta(days=64),
        'min_interval': datetime.timedelta(hours=12),
        'max_interval': datetime.timedelta(days=64),
        'backoff_factor': 2,
        'max_queue_length': None,
        'num_retries': 7,
        'retry_delay': datetime.timedelta(hours=2),
    },
}


def tasks_from_template(template, max_timestamp, num,
                        num_priority=0, priorities=None):
    """Build tasks from template

    """
    def _task_from_template(template, next_run, priority, *args, **kwargs):
        ret = copy.deepcopy(template)
        ret['next_run'] = next_run
        if priority:
            ret['priority'] = priority
        if args:
            ret['arguments']['args'] = list(args)
        if kwargs:
            ret['arguments']['kwargs'] = kwargs
        return ret

    def _pop_priority(priorities):
        if not priorities:
            return None
        for priority, remains in priorities.items():
            if remains > 0:
                priorities[priority] = remains - 1
                return priority
        return None

    if num_priority and priorities:
        priorities = {
            priority: ratio * num_priority
            for priority, ratio in priorities.items()
        }

    tasks = []
    for i in range(num + num_priority):
        priority = _pop_priority(priorities)
        tasks.append(_task_from_template(
            template,
            max_timestamp - datetime.timedelta(microseconds=i),
            priority,
            'argument-%03d' % i,
            **{'kwarg%03d' % i: 'bogus-kwarg'}
        ))
    return tasks



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
from flask import url_for

import swh.scheduler.api.server as server
from swh.scheduler.api.client import RemoteScheduler
from swh.scheduler.tests.test_scheduler import TestScheduler  # noqa

# tests are executed using imported class (TestScheduler) using overloaded
# swh_scheduler fixture below


# the Flask app used as server in these tests
@pytest.fixture
def app(swh_db_scheduler):
    server.scheduler = swh_db_scheduler
    yield server.app


# the RPCClient class used as client used in these tests
@pytest.fixture
def swh_rpc_client_class():
    return RemoteScheduler


@pytest.fixture
def swh_scheduler(swh_rpc_client, app):
    yield swh_rpc_client


def test_site_map(flask_app_client):
    sitemap = flask_app_client.get(url_for('site_map'))
    assert sitemap.headers['Content-Type'] == 'application/json'

    rules = set(x['rule'] for x in sitemap.json)
    # we expect at least these rules
    expected_rules = set('/'+rule for rule in (
        'set_status_tasks', 'create_task_type',
        'get_task_type', 'get_task_types', 'create_tasks', 'disable_tasks',
        'get_tasks', 'search_tasks', 'get_task_runs', 'peek_ready_tasks',
        'grab_ready_tasks', 'schedule_task_run', 'mass_schedule_task_runs',
        'start_task_run', 'end_task_run', 'filter_task_to_archive',
        'delete_archived_tasks', 'get_priority_ratios'))
    assert rules == expected_rules



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Memory Elastic Search backend

"""

import datetime  # noqa serialization purposes
import hashlib
import logging

from ast import literal_eval
from typing import Optional

import psycopg2  # noqa serialization purposes


logger = logging.getLogger(__name__)


class BasicSerializer:
    """For memory elastic search implementation (not for production)

    """
    def __init__(self, *args, **kwargs):
        pass

    def dumps(self, *args, **kwargs):
        return str(*args)


class BasicTransport:
    """For memory elastic search implementation, (not for production)

    """
    def __init__(self, *args, **kwargs):
        self.serializer = BasicSerializer()


class MemoryElasticsearch:
    """Memory Elasticsearch instance (for test purposes)

    Partial implementation oriented towards index storage (and not search)

    For now, its sole client is the scheduler for task archival purposes.

    """
    def __init__(self, *args, **kwargs):
        self.index = {}
        self.mapping = {}
        self.settings = {}
        self.indices = self  # HACK
        self.main_mapping_key: Optional[str] = None
        self.main_settings_key: Optional[str] = None
        self.transport = BasicTransport()

    def create(self, index, **kwargs):
        logger.debug(f'create index {index}')
        logger.debug(f'indices: {self.index}')
        logger.debug(f'mapping: {self.mapping}')
        logger.debug(f'settings: {self.settings}')
        self.index[index] = {
            'status': 'opened',
            'data': {},
            'mapping': self.get_mapping(self.main_mapping_key),
            'settings': self.get_settings(self.main_settings_key),
        }
        logger.debug(f'index {index} created')

    def close(self, index, **kwargs):
        """Close index"""
        idx = self.index.get(index)
        if idx:
            idx['status'] = 'closed'

    def open(self, index, **kwargs):
        """Open index"""
        idx = self.index.get(index)
        if idx:
            idx['status'] = 'opened'

    def bulk(self, body, **kwargs):
        """Bulk insert document in index"""
        assert isinstance(body, str)
        all_data = body.split('\n')
        if all_data[-1] == '':
            all_data = all_data[:-1]  # drop the empty line if any
        ids = []
        # data is sent as tuple (index, data-to-index)
        for i in range(0, len(all_data), 2):
            # The first entry is about the index to use
            # not about a data to index
            # find the index
            index_data = literal_eval(all_data[i])
            idx_name = index_data['index']['_index']
            # associated data to index
            data = all_data[i+1]
            _id = hashlib.sha1(data.encode('utf-8')).hexdigest()
            parsed_data = eval(data)  # for datetime
            self.index[idx_name]['data'][_id] = parsed_data
            ids.append(_id)

        # everything is indexed fine
        return {
            'items': [
                {
                    'index': {
                        'status': 200,
                        '_id': _id,
                    }
                } for _id in ids
            ]
        }

    def mget(self, *args, body, index, **kwargs):
        """Bulk indexed documents retrieval"""
        idx = self.index[index]
        docs = []
        idx_docs = idx['data']
        for _id in body['ids']:
            doc = idx_docs.get(_id)
            if doc:
                d = {
                    'found': True,
                    '_source': doc,
                }
                docs.append(d)
        return {'docs': docs}

    def stats(self, index, **kwargs):
        idx = self.index[index]  # will raise if it does not exist
        if not idx or idx['status'] == 'closed':
            raise ValueError('Closed index')  # simulate issue if index closed

    def exists(self, index, **kwargs):
        return self.index.get(index) is not None

    def put_mapping(self, index, body, **kwargs):
        self.mapping[index] = body
        self.main_mapping_key = index

    def get_mapping(self, index, **kwargs):
        return self.mapping.get(index) or \
            self.index.get(index, {}).get('mapping', {})

    def put_settings(self, index, body, **kwargs):
        self.settings[index] = body
        self.main_settings_key = index

    def get_settings(self, index, **kwargs):
        return self.settings.get(index) or \
            self.index.get(index, {}).get('settings', {})



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import celery.app.task
import click
import logging

from importlib import import_module
from typing import Mapping

from pkg_resources import iter_entry_points
from . import cli


logger = logging.getLogger(__name__)


DEFAULT_TASK_TYPE = {
    'full': {  # for tasks like 'list_xxx_full()'
        'default_interval': '90 days',
        'min_interval': '90 days',
        'max_interval': '90 days',
        'backoff_factor': 1
    },
    '*': {  # value if not suffix matches
        'default_interval': '1 day',
        'min_interval': '1 day',
        'max_interval': '1 day',
        'backoff_factor': 1
    },
}


PLUGIN_WORKER_DESCRIPTIONS = {
    entry_point.name: entry_point
    for entry_point in iter_entry_points('swh.workers')
}


@cli.group('task-type')
@click.pass_context
def task_type(ctx):
    """Manipulate task types."""
    pass


@task_type.command('list')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Verbose mode')
@click.option('--task_type', '-t', multiple=True, default=None,
              help='List task types of given type')
@click.option('--task_name', '-n', multiple=True, default=None,
              help='List task types of given backend task name')
@click.pass_context
def list_task_types(ctx, verbose, task_type, task_name):
    click.echo("Known task types:")
    if verbose:
        tmpl = click.style('{type}: ', bold=True) + '''{backend_name}
  {description}
  interval: {default_interval} [{min_interval}, {max_interval}]
  backoff_factor: {backoff_factor}
  max_queue_length: {max_queue_length}
  num_retries: {num_retries}
  retry_delay: {retry_delay}
'''
    else:
        tmpl = '{type}:\n  {description}'
    for tasktype in sorted(ctx.obj['scheduler'].get_task_types(),
                           key=lambda x: x['type']):
        if task_type and tasktype['type'] not in task_type:
            continue
        if task_name and tasktype['backend_name'] not in task_name:
            continue
        click.echo(tmpl.format(**tasktype))


@task_type.command('register')
@click.option('--plugins', '-p', 'plugins', multiple=True, default=('all', ),
              type=click.Choice(['all'] + list(PLUGIN_WORKER_DESCRIPTIONS)),
              help='Registers task-types for provided plugins. '
                   'Defaults to all')
@click.pass_context
def register_task_types(ctx, plugins):
    """Register missing task-type entries in the scheduler.

    According to declared tasks in each loaded worker (e.g. lister, loader,
    ...) plugins.

    """
    scheduler = ctx.obj['scheduler']

    if plugins == ('all', ):
        plugins = list(PLUGIN_WORKER_DESCRIPTIONS)

    for plugin in plugins:
        entrypoint = PLUGIN_WORKER_DESCRIPTIONS[plugin]
        logger.info('Loading entrypoint for plugin %s', plugin)
        registry_entry = entrypoint.load()()

        for task_module in registry_entry['task_modules']:
            mod = import_module(task_module)
            for task_name in (x for x in dir(mod) if not x.startswith('_')):
                logger.debug('Loading task name %s', task_name)
                taskobj = getattr(mod, task_name)
                if isinstance(taskobj, celery.app.task.Task):
                    tt_name = task_name.replace('_', '-')
                    task_cfg = registry_entry.get('task_types', {}).get(
                        tt_name, {})
                    ensure_task_type(
                        task_module, tt_name, taskobj, task_cfg, scheduler)


def ensure_task_type(
        task_module: str, task_type: str,
        swhtask, task_config: Mapping, scheduler):
    """Ensure a given task-type (for the task_module) exists in the scheduler.

    Args:
        task_module: task module we are currently checking for task type
            consistency
        task_type: the type of the task to check/insert (correspond to
            the 'type' field in the db)
        swhtask (SWHTask): the SWHTask instance the task-type correspond to
        task_config: a dict with specific/overloaded values for the
            task-type to be created
        scheduler: the scheduler object used to access the scheduler db

    """
    for suffix, defaults in DEFAULT_TASK_TYPE.items():
        if task_type.endswith('-' + suffix):
            task_type_dict = defaults.copy()
            break
    else:
        task_type_dict = DEFAULT_TASK_TYPE['*'].copy()

    task_type_dict['type'] = task_type
    task_type_dict['backend_name'] = swhtask.name
    if swhtask.__doc__:
        task_type_dict['description'] = swhtask.__doc__.splitlines()[0]

    task_type_dict.update(task_config)

    current_task_type = scheduler.get_task_type(task_type)
    if current_task_type:
        # Ensure the existing task_type is consistent in the scheduler
        if current_task_type['backend_name'] != task_type_dict['backend_name']:
            logger.warning(
                'Existing task type %s for module %s has a '
                'different backend name than current '
                'code version provides (%s vs. %s)',
                task_type,
                task_module,
                current_task_type['backend_name'],
                task_type_dict['backend_name'],
            )
    else:
        logger.info('Create task type %s in scheduler', task_type)
        logger.debug('  %s', task_type_dict)
        scheduler.create_task_type(task_type_dict)


@task_type.command('add')
@click.argument('type', required=True)
@click.argument('task-name', required=True)
@click.argument('description', required=True)
@click.option('--default-interval', '-i', default='90 days',
              help='Default interval ("90 days" by default)')
@click.option('--min-interval', default=None,
              help='Minimum interval (default interval if not set)')
@click.option('--max-interval', '-i', default=None,
              help='Maximal interval (default interval if not set)')
@click.option('--backoff-factor', '-f', type=float, default=1,
              help='Backoff factor')
@click.pass_context
def add_task_type(ctx, type, task_name, description,
                  default_interval, min_interval, max_interval,
                  backoff_factor):
    """Create a new task type
    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')
    task_type = dict(
        type=type,
        backend_name=task_name,
        description=description,
        default_interval=default_interval,
        min_interval=min_interval,
        max_interval=max_interval,
        backoff_factor=backoff_factor,
        max_queue_length=None,
        num_retries=None,
        retry_delay=None,
        )
    scheduler.create_task_type(task_type)
    click.echo('OK')



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import os
import pytest
import glob
from datetime import timedelta
import pkg_resources

from swh.core.utils import numfile_sortkey as sortkey
from swh.scheduler import get_scheduler
from swh.scheduler.tests import SQL_DIR


# make sure we are not fooled by CELERY_ config environment vars
for var in [x for x in os.environ.keys() if x.startswith('CELERY')]:
    os.environ.pop(var)


# test_cli tests depends on a en/C locale, so ensure it
os.environ['LC_ALL'] = 'C.UTF-8'

DUMP_FILES = os.path.join(SQL_DIR, '*.sql')

# celery tasks for testing purpose; tasks themselves should be
# in swh/scheduler/tests/tasks.py
TASK_NAMES = ['ping', 'multiping', 'add', 'error', 'echo']


@pytest.fixture(scope='session')
def celery_enable_logging():
    return True


@pytest.fixture(scope='session')
def celery_includes():
    task_modules = [
        'swh.scheduler.tests.tasks',
    ]
    for entrypoint in pkg_resources.iter_entry_points('swh.workers'):
        task_modules.extend(entrypoint.load()().get('task_modules', []))
    return task_modules


@pytest.fixture(scope='session')
def celery_parameters():
    return {
        'task_cls': 'swh.scheduler.task:SWHTask',
        }


@pytest.fixture(scope='session')
def celery_config():
    return {
        'accept_content': ['application/x-msgpack', 'application/json'],
        'task_serializer': 'msgpack',
        'result_serializer': 'json',
        }


# use the celery_session_app fixture to monkeypatch the 'main'
# swh.scheduler.celery_backend.config.app Celery application
# with the test application
@pytest.fixture(scope='session')
def swh_app(celery_session_app):
    from swh.scheduler.celery_backend import config
    config.app = celery_session_app
    yield celery_session_app


@pytest.fixture
def swh_scheduler_config(request, postgresql):
    scheduler_config = {
        'db': postgresql.dsn,
    }

    all_dump_files = sorted(glob.glob(DUMP_FILES), key=sortkey)

    cursor = postgresql.cursor()
    for fname in all_dump_files:
        with open(fname) as fobj:
            cursor.execute(fobj.read())
    postgresql.commit()

    return scheduler_config


@pytest.fixture
def swh_scheduler(swh_scheduler_config):
    scheduler = get_scheduler('local', swh_scheduler_config)
    for taskname in TASK_NAMES:
        scheduler.create_task_type({
            'type': 'swh-test-{}'.format(taskname),
            'description': 'The {} testing task'.format(taskname),
            'backend_name': 'swh.scheduler.tests.tasks.{}'.format(taskname),
            'default_interval': timedelta(days=1),
            'min_interval': timedelta(hours=6),
            'max_interval': timedelta(days=12),
        })

    return scheduler


# this alias is used to be able to easily instantiate a db-backed Scheduler
# eg. for the RPC client/server test suite.
swh_db_scheduler = swh_scheduler
